#include <climits>
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <vector>
#include <utility>
#include <ctime>
#include <cmath>
#include <cassert>
#include <algorithm>
#include <iostream>
#include "sudoku.h"

using namespace std;

#define EPS 1E-10
#define mem(x) memset(x, 0, sizeof(x))

unsigned long xor_shift(){ 
    static unsigned long long x = 123456789;
    static unsigned long long y = 362436069;
    static unsigned long long z = 521288629;
    static unsigned long long w = 88675123; 
    unsigned long t;
    t = (x ^ (x << 11));
    x = y; y = z; z = w;
    return (w = (w ^ (w >> 19)) ^ (t ^ (t >> 8))); 
}

inline double rand01() {
    return xor_shift() * 1.0 / ULLONG_MAX;
}

// [low, high)
inline int rand_range(int low, int high) {
    assert(high > low);
    return (int) (low + xor_shift() % (high - low));
}

Sudoku::Sudoku() {}

Sudoku::Sudoku(vector<vector<int> > board) {
    size = board.size();
    cells.resize(size, vector<int>(size));
    for (int i = 0; i < size; ++i) {
        for (int j = 0; j < size; ++j) {
            cells[i][j] = board[i][j];
        }
    }
}

bool Sudoku::is_solved() {
    int cnt[10];

    /* check rows */
    for (int i = 0; i < size; ++i) {
        mem(cnt);
        for (int j = 0; j < size; ++j) {
            cnt[cells[i][j]]++;
        }
        for (int k = 1; k < 10; ++k) {
            if (cnt[k] != 1) return false;
        }
    }

    /* check columns */
    for (int j = 0; j < size; ++j) {
        mem(cnt);
        for (int i = 0; i < size; ++i) {
            cnt[cells[i][j]]++;
        }
        for (int k = 1; k < 10; ++k) {
            if (cnt[k] != 1) return false;
        }
    }

    /* check 3x3 sub-squares*/
    for (int i = 0; i < size; i += 3) {
        for (int j = 0; j < size; j += 3) {
            mem(cnt);
            for (int k = 0; k < 3; ++k) {
                for (int l = 0; l < 3; ++l) {
                    cnt[cells[i + k][j + l]]++;
                }
            }
            for (int k = 1; k < 10; ++k) {
                if (cnt[k] != 1) return false;
            }
        }
    }
    return true;
}

vector<vector<int> > Sudoku::get_result() {
    return cells;
}

Sudoku* solve(Sudoku* sdk) {
    GA ga = GA(sdk->get_result(), 1000, 100000);
    cout << "Starting solve by GA" << endl;
    Sudoku* res = ga.simulate();
	cout << "Run time = " << (clock() * 1.0 / CLOCKS_PER_SEC) << endl;
    if (res->is_solved()) cout << "SOLVED OK!" << endl;
    else cout << "WRONG ANSWER" << endl;
    return res;
}

Individual::Individual() {}

Individual::Individual(vector<vector<int> > _gene) {
	gene = _gene;
}

double Individual::calculate_fitness() {
	double fitness_rows = 0, fitness_columns = 0, fitness_squares = 0;
	int cnt[10], tmp;
	
	// calculate fitness of rows
	for (int i = 0; i < 9; ++i) {
		mem(cnt);
		for (int j = 0; j < 9; ++j) {
			cnt[gene[i][j]] = 1;
		}
		tmp = 0;
		for (int j = 1; j < 10; ++j) {
			if (cnt[j]) tmp++;
		}
		fitness_rows += 1.0 / (10 - tmp);
	}
	fitness_rows /= 9;
	
	// calculate fitness of columns
	for (int j = 0; j < 9; ++j) {
		mem(cnt);
		for (int i = 0; i < 9; ++i) {
			cnt[gene[i][j]] = 1;
		}
		tmp = 0;
		for (int i = 1; i < 10; ++i) {
			if (cnt[i]) tmp++;
		}
		fitness_columns += 1.0 / (10 - tmp);
	}
	fitness_columns /= 9;

	
	// calculate fitness of 3x3 sub-squares
	for (int i = 0; i < 9; i += 3) {
		for (int j = 0; j < 9; j += 3) {
			mem(cnt);
			for (int x = i; x < i + 3; ++x) {
				for (int y = j; y < j + 3; ++y) {
					cnt[gene[x][y]] = 1;
				}
			}
			tmp = 0;
			for (int k = 1; k < 10; ++k) {
				if (cnt[k]) tmp++;
			}
			fitness_squares += 1.0 / (10 - tmp);
		}
	}
	fitness_squares /= 9;
	fitness = fitness_rows * fitness_columns * fitness_squares;

	return fitness;
}

GA::GA() {}

GA::GA(vector<vector<int> > _board, int _num_individuals, int _iterations) {
	board = _board;
	num_individuals = _num_individuals;
	iterations = _iterations;
	crossover_rate = 0.75;
	best_fitness = 0;
	prob.resize(num_individuals);

	// determine all empty cells
	for (int i = 0; i < 9; ++i) {
		for (int j = 0; j < 9; ++j) {
			if (board[i][j] == 0) flex.push_back(9 * i + j);
		}
	}
	if (flex.size()) mutation_rate = min(0.5, 1.2 / flex.size());
	else mutation_rate = 0.01;

	int cnt[10];
	for (int i = 0; i < flex.size(); ++i) {
    	int i0 = flex[i] / 9, j0 = flex[i] % 9;
    	mem(cnt);
    	for (int k = 0; k < 9; ++k) {
    		cnt[board[i0][k]]++;
    	}
    	for (int k = 0; k < 9; ++k) {
    		cnt[board[k][j0]]++;
    	}
    	int x = i0 / 3, y = j0 / 3;
    	for (int ii = x * 3; ii < x * 3 + 3; ++ii) {
    		for (int jj = y * 3; jj < y * 3 + 3; ++jj) {
    			cnt[board[ii][jj]]++;
    		}
    	}
    	for (int k = 1; k < 10; ++k) {
    		if (cnt[k] == 0) candidates[flex[i]].push_back(k);
    	}
    }
}

void GA::initialize() {
    vector<vector<int> > tmp(board);
    for (int i = 0; i < num_individuals; ++i) {
    	for (int j = 0; j < flex.size(); ++j) {
    		int i0 = flex[j] / 9, j0 = flex[j] % 9;
    		tmp[i0][j0] = candidates[flex[j]][rand_range(0, candidates[flex[j]].size())];
    	}
        current.push_back(Individual(tmp));
    }
}

bool comparator(const Individual& id1, const Individual& id2) {
	return id1.fitness > id2.fitness;
}

void GA::mutate(Individual& par) {
	for (int i = 0; i < flex.size(); ++i) {
		if (rand01() <= mutation_rate) {
			par.gene[flex[i] / 9][flex[i] % 9] = candidates[flex[i]][rand_range(0, candidates[flex[i]].size())];
		}
	}
}

void GA::cross_over(Individual& par1, Individual& par2) {
	int lb = rand_range(0, 81);
	int rb = rand_range(0, 81);
	if (lb > rb) swap(lb, rb);
	for (int k = lb; k <= rb; ++k) {
		int i = k / 9, j = k % 9;
		swap(par1.gene[i][j], par2.gene[i][j]);
	}
}

void GA::next_generation() {
    double sum = 0;

    for (int i = 0; i < num_individuals; ++i) {
    	current[i].calculate_fitness();
    }
    
    sort(current.begin(), current.end(), comparator);
    if (best_fitness < current[0].fitness) {
    	best_fitness = current[0].fitness;
    	best_individual = current[0];
    }

    // scaling rank n individual's fitness to 1/sqrt(n)
    for (int i = 0; i < num_individuals; ++i) {
        prob[i] = current[i].fitness;
        sum += prob[i];
    }
    
    for (int i = 0; i < num_individuals; ++i) {
        prob[i] /= sum;
    }
    
    vector<Individual> next_gen;
    int num_crossover = (int) (num_individuals * crossover_rate);
	if (num_crossover & 1) ++num_crossover;
    int num_elites = num_individuals - num_crossover;
    // choose some best elite children to next generation
    for (int i = 0; i < num_elites; ++i) {
    	next_gen.push_back(current[i]);
    }
    
    // cross over parent by probability selection
    Individual par[2];
    for (int i = 0; i < (num_crossover >> 1); ++i) {
    	int pre = -1;
    	bool duplicate = 0;
        for (int j = 0; j < 2;) {
            double tot = rand01();
            for (int k = 0; k < num_individuals; ++k) {
                tot -= prob[k];
                if (tot < EPS) {
                    par[j] = current[k];
                    if (j == 0) pre = k;
                    else if (k == pre) duplicate = 1;
                    break;
                }
            }
            if (!duplicate) j++;
            else duplicate = 0;
        }
        // cross parent to create offsprings
        cross_over(par[0], par[1]);
        // add two children to next generation
        next_gen.push_back(par[0]);
        next_gen.push_back(par[1]);
    }
    
    // mutation
    for (int i = 0; i < num_individuals; ++i) {
        mutate(next_gen[i]);
    }
    
    // go to next generation
    current = next_gen;
}

Sudoku* GA::simulate() {
	initialize();
	for (int it = 0; it < iterations; ++it) {
		next_generation();
		if (it % 100 == 0) {
			cout << "[" << it << "] " << "best_fitness = " << best_fitness << ",cur_fitness = " << current[0].fitness << endl;
			for (int i = 0; i < 9; ++i) for (int j = 0; j < 9; ++j) {
				cout << best_individual.gene[i][j] << (j == 8 ? '\n' : ' ');
			}
		}
		if (fabs(best_fitness - 1) < EPS) break;
	}
	for (int i = 0; i < 9; ++i) {
		for (int j = 0; j < 9; ++j) {
			cout << best_individual.gene[i][j] << (j == 8 ? '\n' : ' ');
		}
	}
	return new Sudoku(best_individual.gene);
}

