#include <cmath>
#include <ctime>
#include <cstring>
#include <climits>
#include <iostream>
#include <cstdlib>
#include <utility>
#include <cstdio>
#include <vector>
#include <cassert>
#include "sudoku.h"

using namespace std;

#define N   9
#define N2  (N * N)
#define dt  0.1
#define T0  0.5
#define MAX_ITER  200000000
#define EPS  1E-4
#define BOUND 15
#define mem(x) memset(x, 0, sizeof(x))

inline double pow2(double x) {
    return x * x;
}

inline double sigmoid(double u, double T) {
    return (1 + tanh(u / T)) / 2;
}

inline int HASH(int i, int j, int k) {
    return N2 * i + N * j + k;
}

unsigned long xor_shift(){ 
    static unsigned long long x = 123456789;
    static unsigned long long y = 362436069;
    static unsigned long long z = 521288629;
    static unsigned long long w = 88675123; 
    unsigned long t;
    t = (x ^ (x << 11));
    x = y; y = z; z = w;
    return (w = (w ^ (w >> 19)) ^ (t ^ (t >> 8))); 
} 

inline double rand01() {
    return xor_shift() * 1.0 / ULLONG_MAX;
}

// [low, high)
inline int rand_range(int low, int high) {
    assert(high > low);
    return (int) (low + xor_shift() % (high - low));
}

Sudoku::Sudoku() {}

Sudoku::Sudoku(vector<vector<int> > board) {
    size = board.size();
    cells.resize(size, vector<int>(size));
    for (int i = 0; i < size; ++i) {
        for (int j = 0; j < size; ++j) {
            cells[i][j] = board[i][j];
        }
    }
}

bool Sudoku::is_solved() {
    int cnt[10];

    /* check rows */
    for (int i = 0; i < size; ++i) {
        mem(cnt);
        for (int j = 0; j < size; ++j) {
            cnt[cells[i][j]]++;
        }
        for (int k = 1; k < 10; ++k) {
            if (cnt[k] != 1) return false;
        }
    }

    /* check columns */
    for (int j = 0; j < size; ++j) {
        mem(cnt);
        for (int i = 0; i < size; ++i) {
            cnt[cells[i][j]]++;
        }
        for (int k = 1; k < 10; ++k) {
            if (cnt[k] != 1) return false;
        }
    }

    /* check 3x3 sub-squares*/
    for (int i = 0; i < size; i += 3) {
        for (int j = 0; j < size; j += 3) {
            mem(cnt);
            for (int k = 0; k < 3; ++k) {
                for (int l = 0; l < 3; ++l) {
                    cnt[cells[i + k][j + l]]++;
                }
            }
            for (int k = 1; k < 10; ++k) {
                if (cnt[k] != 1) return false;
            }
        }
    }
    return true;
}

vector<vector<int> > Sudoku::get_result() {
    return cells;
}

Neuron::Neuron() {}

Neuron::Neuron(double _u, double _v) {
    u = _u; v = _v;
}

HopfieldNN::HopfieldNN() {}

HopfieldNN::HopfieldNN(vector<vector<int> > board) {
    size = board.size();
    num_neuron = size * size * size;
    neurons.resize(num_neuron);
    mark.assign(num_neuron, 0);
    T = T0;
    for (int i = 0; i < num_neuron; ++i) {
        neurons[i].u = 0.01 * (rand01() - 0.5);
    }
    for (int i = 0; i < size; ++i) {
        for (int j = 0; j < size; ++j) {
            if (board[i][j]) {
                for (int k = 0; k < size; ++k) {
                    mark[HASH(i, j, k)] = 1;
                    neurons[HASH(i, j, k)].u = -BOUND;
                }
                neurons[HASH(i, j, board[i][j] - 1)].u = BOUND;
                for (int l = 0; l < size; ++l) {
                    if (board[l][j] == 0) {
                        mark[HASH(l, j, board[i][j] - 1)] = 1;
                        neurons[HASH(l, j, board[i][j] - 1)].u = -BOUND;
                    }
                    if (board[i][l] == 0) {
                        mark[HASH(i, l, board[i][j] - 1)] = 1;
                        neurons[HASH(i, l, board[i][j] - 1)].u = -BOUND;
                    }
                }
                int i0 = i / 3, j0 = j / 3;
                for (int x = i0 * 3; x < i0 * 3 + 3; ++x) {
                    for (int y = j0 * 3; y < j0 * 3 + 3; ++y) {
                        if (board[x][y] == 0) {
                            mark[HASH(x, y, board[i][j] - 1)] = 1;
                            neurons[HASH(x, y, board[i][j] - 1)].u = -BOUND;
                        }
                    }
                }
            }
        }
    }

    for (int i = 0; i < num_neuron; ++i) {
        neurons[i].v = sigmoid(neurons[i].u, T);
        if (!mark[i]) flex.push_back(i);
    }
}

double HopfieldNN::calc_energy() {
    double ans = 0, tmp;

    for (int x = 0; x < size; ++x) {
        for (int y = 0; y < size; ++y) {
            tmp = 0;
            for (int z = 0; z < size; ++z) {
                tmp += neurons[HASH(x, y, z)].v;
            }
            ans += pow2(tmp - 1);
		}
	}
	for (int x = 0; x < size; ++x) {
        for (int z = 0; z < size; ++z) {
            tmp = 0;
            for (int y = 0; y < size; ++y) {
                tmp += neurons[HASH(x, y, z)].v;
            }
            ans += pow2(tmp - 1);
		}
	}
	for (int y = 0; y < size; ++y) {
        for (int z = 0; z < size; ++z) {
            tmp = 0;
            for (int x = 0; x < size; ++x) {
                tmp += neurons[HASH(x, y, z)].v;
            }
            ans += pow2(tmp - 1);
        }
    }
    for (int i = 0; i < 3; ++i) for (int j = 0; j < 3; ++j) {
        for (int z = 0; z < size; ++z) {
            tmp = 0;
            for (int x = i * 3; x < i * 3 + 3; ++x) {
                for (int y = j * 3; y < j * 3 + 3; ++y) {
                    tmp += neurons[HASH(x, y, z)].v;
                }
            }
            ans += pow2(tmp - 1);
        }
    }
    ans /= 2;
    return ans;
}

void HopfieldNN::update(int idx) {
    double tmp;
    int i0, j0;
    int node = flex[idx];

    int i = node / N2, j = (node / N) % N, k = node % N;
    tmp = 4;
    for (int l = 0; l < size; ++l) {
        tmp -= neurons[HASH(l, j, k)].v + neurons[HASH(i, l, k)].v + neurons[HASH(i, j, l)].v;
    }
    i0 = i / 3, j0 = j / 3;
    for (int x = i0 * 3; x < i0 * 3 + 3; ++x) {
        for (int y = j0 * 3; y < j0 * 3 + 3; ++y) {
            tmp -= neurons[HASH(x, y, k)].v;
        }
    }
    neurons[node].u += tmp * dt;
    neurons[node].v = sigmoid(neurons[node].u, T);
}

Sudoku* HopfieldNN::simulate() {
    int iter = 0;
    double energy = 1;
    srand(time(NULL));
    int sz = flex.size();
    if (sz) {
        while (++iter < MAX_ITER) {
            energy = calc_energy();
            if (iter % 10000 == 0) cout << "Iterations = " << iter << ", Energy = " << energy << endl;
            if (energy < EPS) {
				bool good = 1;
				for (int i = 0; i < size; ++i) {
        			for (int j = 0; j < size; ++j) {
            			double maxi = -1;
            			int val = 0;
            			for (int k = 0; k < size; ++k) {
                			if (neurons[HASH(i, j, k)].v >= maxi) {
                    			maxi = neurons[HASH(i, j, k)].v;
                    			val = k;
                			}
						}
						if (maxi < 0.5) {
							good = 0;
							neurons[HASH(i, j, val)].u = -BOUND;
							neurons[HASH(i, j, val)].v = sigmoid(-BOUND, T);
						}
					}
				}
				if (good) break;
            }
            update(rand_range(0, sz));
        }
        cout << "------------------------" << endl;
        cout << "Iterations = " << iter << ", Final energy = " << energy << endl;
    }

    vector<vector<int> > board(size, vector<int>(size, 0));
    for (int i = 0; i < size; ++i) {
        for (int j = 0; j < size; ++j) {
            double maxi = -1;
            int val = 0;
            for (int k = 0; k < size; ++k) {
                if (neurons[HASH(i, j, k)].v >= maxi) {
                    maxi = neurons[HASH(i, j, k)].v;
                    val = k + 1;
                }
            }
			cout << maxi << (j == 8 ? '\n' : ' ');
            board[i][j] = val;
        }
    }
    return new Sudoku(board);
}

Sudoku* solve(Sudoku* sdk) {
    HopfieldNN hopNN = HopfieldNN(sdk->get_result());
    Sudoku* res = hopNN.simulate();
    if (res->is_solved()) cout << "SOLVED OK!" << endl;
    else cout << "WRONG ANSWER" << endl;
    return res;
}
